# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: keras/protobuf/saved_metadata.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from keras.protobuf import versions_pb2 as keras_dot_protobuf_dot_versions__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#keras/protobuf/saved_metadata.proto\x12\x1dthird_party.py.keras.protobuf\x1a\x1dkeras/protobuf/versions.proto\"J\n\rSavedMetadata\x12\x39\n\x05nodes\x18\x01 \x03(\x0b\x32*.third_party.py.keras.protobuf.SavedObject\"\x99\x01\n\x0bSavedObject\x12\x0f\n\x07node_id\x18\x02 \x01(\x05\x12\x11\n\tnode_path\x18\x03 \x01(\t\x12\x12\n\nidentifier\x18\x04 \x01(\t\x12\x10\n\x08metadata\x18\x05 \x01(\t\x12:\n\x07version\x18\x06 \x01(\x0b\x32).third_party.py.keras.protobuf.VersionDefJ\x04\x08\x01\x10\x02\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'keras.protobuf.saved_metadata_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SAVEDMETADATA._serialized_start=101
  _SAVEDMETADATA._serialized_end=175
  _SAVEDOBJECT._serialized_start=178
  _SAVEDOBJECT._serialized_end=331
# @@protoc_insertion_point(module_scope)
